using System;
using System.Reflection;
using FluentAssertions;
using Xunit;

namespace DBurstin.GrowingIDisposable
{
    public class TestThatResourceOwner
    {
        [Fact]
        public void ReleasesUnmanagedResourcesWhenFinalized()
        {
            var spy = new ResourceSpy();
            spy.MimicFinalizer();
            spy.UnmanagedResourcesHaveBeenReleased.Should().BeTrue();
            GC.SuppressFinalize(spy);
        }

        [Fact]
        public void ReleasesUnmanagedResourcesFromDispose()
        {
            var spy = new ResourceSpy();
            spy.Dispose();
            spy.UnmanagedResourcesHaveBeenReleased.Should().BeTrue();
        }

        [Fact]
        public void ImplementsIDisposable()
        {
            var resourceOwner = new ResourceOwner() as IDisposable;
            resourceOwner.Should().NotBeNull();
        }

        [Fact]
        public void ProvidesAVirtualProtectedDisposeMethodWithABooleanParameter()
        {
            var methodInfo = typeof(ResourceOwner)
                                .GetMethod("Dispose",
                                    BindingFlags.Instance | BindingFlags.NonPublic | BindingFlags.Public,
                                    null,
                                    new[] { typeof(bool) },
                                    null);

            Assert.NotNull(methodInfo); //There is a Dispose(bool) method
            methodInfo.IsFamily.Should().BeTrue("the Dispose(bool) method is protected");
            methodInfo.IsVirtual.Should().BeTrue("the Dispose(bool) method is virtual");
        }

        [Fact]
        public void CallsDisposeWithTrueFromDispose()
        {
            var spy = new ResourceSpy();
            spy.Dispose();
            spy.DisposeWasCalled(true).Should().BeTrue();
        }

        [Fact]
        public void CallsDisposeWithFalseFromFinalizer()
        {
            var spy = new ResourceSpy();
            spy.MimicFinalizer();
            spy.DisposeWasCalled(false).Should().BeTrue();
            GC.SuppressFinalize(spy);
        }

        [Fact]
        public void HandlesDisposeBeingCalledMultipleTimes()
        {
            var resourceOwner = new ResourceOwner();
            resourceOwner.Dispose();
            resourceOwner.Dispose();
        }

        [Fact]
        public void DisposesDisposableResourcesFromDispose()
        {
            var spy = new ResourceSpy();
            spy.Dispose();
            spy.DisposableResourcesHaveBeenDisposed.Should().BeTrue();
        }

        [Fact]
        public void DoesNotDisposeDisposableResourcesFromFinalizer()
        {
            var spy = new ResourceSpy();
            spy.MimicFinalizer();
            spy.DisposableResourcesHaveBeenDisposed.Should().BeFalse();
            GC.SuppressFinalize(spy);
        }

        [Fact]
        public void ThrowsObjectDisposedExceptionIfDisposedResourcesAccessed()
        {
            var resourceOwner = new ResourceOwner();

            resourceOwner.Invoking(r => r.PublicMethodThatUsesFinalizableResources())
                         .ShouldNotThrow<ObjectDisposedException>();

            resourceOwner.Dispose();

            resourceOwner.Invoking(r => r.PublicMethodThatUsesFinalizableResources())
                         .ShouldThrow<ObjectDisposedException>();

        }
    }
}
